/*******************************************************************************

"FreePastry" Peer-to-Peer Application Development Substrate

Copyright 2002-2007, Rice University. Copyright 2006-2007, Max Planck Institute 
for Software Systems.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither the name of Rice  University (RICE), Max Planck Institute for Software 
Systems (MPI-SWS) nor the names of its contributors may be used to endorse or 
promote products derived from this software without specific prior written 
permission.

This software is provided by RICE, MPI-SWS and the contributors on an "as is" 
basis, without any representations or warranties of any kind, express or implied 
including, but not limited to, representations or warranties of 
non-infringement, merchantability or fitness for a particular purpose. In no 
event shall RICE, MPI-SWS or contributors be liable for any direct, indirect, 
incidental, special, exemplary, or consequential damages (including, but not 
limited to, procurement of substitute goods or services; loss of use, data, or 
profits; or business interruption) however caused and on any theory of 
liability, whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even if 
advised of the possibility of such damage.

*******************************************************************************/ 
/*
 * Created on Nov 8, 2005
 */
package rice.pastry.direct;

import java.util.*;

import org.mpisws.p2p.transport.TransportLayer;
import org.mpisws.p2p.transport.liveness.LivenessListener;
import org.mpisws.p2p.transport.proximity.ProximityListener;

import rice.environment.Environment;
import rice.environment.logging.Logger;
import rice.environment.params.Parameters;
import rice.environment.random.RandomSource;
import rice.environment.random.simple.SimpleRandomSource;
import rice.environment.time.TimeSource;
import rice.environment.time.simulated.DirectTimeSource;
import rice.p2p.commonapi.Cancellable;
import rice.p2p.commonapi.CancellableTask;
import rice.pastry.*;
import rice.pastry.messaging.Message;
import rice.pastry.transport.NodeHandleAdapter;
import rice.selector.SelectorManager;
import rice.selector.TimerTask;

public class BasicNetworkSimulator<Identifier, MessageType> implements 
    GenericNetworkSimulator<Identifier, MessageType> {
  class Tupel {
    Identifier i;
    TransportLayer<Identifier, MessageType> tl;
    NodeRecord record;
    
    public Tupel(Identifier i, TransportLayer<Identifier, MessageType> tl, NodeRecord record) {
      super();
      this.i = i;
      this.tl = tl;
      this.record = record;
    }
  }
  
  /**
   * This maps to the next highest transport layer
   */
  Map<Identifier, Tupel> nodes = 
    new HashMap<Identifier, Tupel>();

  Environment environment;

  TimeSource timeSource;

  // true if we are responsible for incrementing the time
  private boolean isDirectTimeSource = false;
  

  protected Logger logger;

  protected RandomSource random;

  protected int MIN_DELAY = 1;
  
  protected SelectorManager manager;
    
  protected final int maxDiameter;
  protected final int minDelay;
  
  public BasicNetworkSimulator(Environment env, RandomSource random) {
    this.environment = env;
    manager = environment.getSelectorManager();
    manager.useLoopListeners(false);
    Parameters params = env.getParameters();
    maxDiameter = params.getInt("pastry_direct_max_diameter");
    minDelay = params.getInt("pastry_direct_min_delay");
    this.random = random;
    this.logger = env.getLogManager().getLogger(getClass(), null);
//    try {
    timeSource = env.getTimeSource();
//    } catch (ClassCastException cce) {
//      throw new IllegalArgumentException(
//          "env.getTimeSource() must return a DirectTimeSource instead of a "
//              + env.getTimeSource().getClass().getName());
//    }
    if (timeSource instanceof DirectTimeSource)
      isDirectTimeSource = true;
    manager.setSelect(false);
    start();
  }

  boolean running = false; // Invariant: only modified on the selector

  public void start() {
    // this makes things single threaded
    manager.invoke(new Runnable() {      
      public void run() {
        if (running) return;
        running = true;
        manager.invoke(new Runnable() {    
          public void run() {
            if (!running) return;            
            try {
            if (!simulate()) {
              synchronized(manager) {
                try {
                  manager.wait(100); // must wait on the real clock, because the simulated clock can only be advanced by simulate()
                } catch (InterruptedException ie) {
                  logger.logException("BasicNetworkSimulator interrupted.",ie); 
                }
              }
            }
            // re-invoke the simulation task
            manager.invoke(this);
            } catch (InterruptedException ie) {
              if (logger.level <= Logger.SEVERE) logger.logException("BasicNetworkSimulator.start()",ie); 
              stop();
            }
          }
        });
      }
    });
  }
    
  public void stop() {
    manager.invoke(new Runnable() {      
      public void run() {
        running = false;
      }
    });
  }
  
  private void addTask(TimerTask dtt) {
    if (logger.level <= Logger.FINE) logger.log("addTask("+dtt+")");
//    System.out.println("addTask("+dtt+")");
//    synchronized(taskQueue) {
//      taskQueue.add(dtt);
//    }
    manager.getTimer().schedule(dtt);
//    start();
//    if (!manager.isSelectorThread()) Thread.yield();
  }
  
  public CancellableTask enqueueDelivery(Delivery d, int delay) {
    long time = timeSource.currentTimeMillis()+delay;
    if (logger.level <= Logger.FINE)      
      logger.log("BNS: enqueueDelivery " + d+":"+time);
    DeliveryTimerTask dtt = null;
    dtt = new DeliveryTimerTask(d, time, d.getSeq());
    addTask(dtt);
    return dtt;
  }
  
  /**
   * node should always be a local node, because this will be delivered instantly
   */
  public Cancellable deliverMessage(MessageType msg, Identifier node, Identifier from) {
    return deliverMessage(msg, node, from, 0);
  }

  public Cancellable deliverMessage(MessageType msg, Identifier node, Identifier from,
      int delay) {
    return deliverMessage(msg, node, from, delay, 0);
  }

  public Cancellable deliverMessageFixedRate(MessageType msg,
      Identifier node, Identifier from, int delay, int period) {
    return deliverMessage(msg, node, from, delay, period);
  }
  
  public Cancellable deliverMessage(MessageType msg, Identifier node, Identifier from, int delay, int period) {
    if (logger.level <= Logger.FINE)
      logger.log("BNS: deliver " + msg + " to " + node);
    DirectTimerTask dtt = null;
    
    if (from == null || isAlive(from)) {
      MessageDelivery<Identifier, MessageType> md = new MessageDelivery<Identifier, MessageType>(msg, node, from, null, this);
      dtt = new DirectTimerTask(md, timeSource.currentTimeMillis()+ delay,period);
      addTask(dtt);
    }
    return dtt;
  }

  
  // System Time is the system clock
  // Sim Time is the simulated clock
  long maxSpeedRequestSystemTime = 0;
  long maxSpeedRequestSimTime = 0;
  float maxSpeed = 0.0f;
  
  /**
   * This is a guardian for printing the "Invalid TimeSource" warning.  So that it is only printed once.
   */
  boolean printedDirectTimeSourceWarning = false;
  
  public void setMaxSpeed(float speed) {
    if (!isDirectTimeSource) {
      if (!printedDirectTimeSourceWarning) {
        if (logger.level <= Logger.WARNING) logger.log("Invalid TimeSource for setMaxSpeed()/setFullSpeed().  Use Environment.directEnvironment() to construct your Environment.");
        printedDirectTimeSourceWarning = true;
      }
    }
    maxSpeedRequestSystemTime = System.currentTimeMillis();
    maxSpeedRequestSimTime = timeSource.currentTimeMillis();
    maxSpeed = speed;
  }
  
  public void setFullSpeed() {
    setMaxSpeed(-1.0f);
  }
  
  /**
   * Delivers 1 message. Will advance the clock if necessary.
   * 
   * If there is a message in the queue, deliver that and return true. If there
   * is a message in the taskQueue, update the clock if necessary, deliver that,
   * then return true. If both are empty, return false;
   */  
  private boolean simulate() throws InterruptedException {
    if (!isDirectTimeSource) return true;
    if (!environment.getSelectorManager().isSelectorThread()) throw new RuntimeException("Must be on selector thread");
    synchronized(manager) { // so we can wait on it, and so the clock and nextExecution don't change
      
      long scheduledExecutionTime = manager.getNextTaskExecutionTime();
      if (scheduledExecutionTime < 0) {
        if (logger.level <= Logger.FINE) logger.log("taskQueue is empty");
        return false;
      }
      
      if (scheduledExecutionTime > timeSource.currentTimeMillis()) {
        long newSimTime = scheduledExecutionTime;
        if (maxSpeed > 0) {
          long sysTime = System.currentTimeMillis();
          long sysTimeDiff = sysTime-maxSpeedRequestSystemTime;
          
          long maxSimTime = (long)(maxSpeedRequestSimTime+(sysTimeDiff*maxSpeed));
          
          if (maxSimTime < newSimTime) {
            // we need to throttle
            long neededSysDelay = (long)((newSimTime-maxSimTime)/maxSpeed);
//            System.out.println("Waiting for "+neededSysDelay);
            if (neededSysDelay >= 1) {
              manager.wait(neededSysDelay);          
              long now = System.currentTimeMillis();
              long delay = now-sysTime;
  //            System.out.println("Woke up after "+delay);
              if (delay < neededSysDelay) return true;
            }
          }
        }
          
        if (logger.level <= Logger.FINER) logger.log("the time is now "+newSimTime);              
        ((DirectTimeSource)timeSource).setTime(newSimTime);      
      }
    } // synchronized(manager)
    
//    TimerTask task;
//    synchronized(taskQueue) {
//      // take a task from the taskQueue
//      if (taskQueue.isEmpty()) {
//        if (logger.level <= Logger.FINE) logger.log("taskQueue is empty");
//        return false;
//      }
//      task = (TimerTask) taskQueue.first();
//      if (logger.level <= Logger.FINE) logger.log("simulate():"+task);
//      taskQueue.remove(task);
//    }
//      // increment the clock if needed
//      if (task.scheduledExecutionTime() > timeSource.currentTimeMillis()) {
//        if (logger.level <= Logger.FINER) logger.log("the time is now "+task.scheduledExecutionTime());        
//        timeSource.setTime(task.scheduledExecutionTime());
//      }
//  
//  
//      if (task.execute(timeSource)) {
//        addTask(task);
//      }    
      
      
      return true;
  }

  /**
   * set the liveliness of a NodeId
   * 
   * @param nid the NodeId being set
   * @param alive the value being set
   */
//  public void destroy(TLPastryNode node) {
//    node.destroy();
//    // NodeRecord nr = (NodeRecord) nodeMap.get(nid);
//    //
//    // if (nr == null) {
//    // throw new Error("setting node alive for unknown node");
//    // }
//    //
//    // if (nr.alive != alive) {
//    // nr.alive = alive;
//    //
//    // DirectNodeHandle[] handles = (DirectNodeHandle[]) nr.handles.toArray(new
//    // DirectNodeHandle[0]);
//    //
//    // for (int i = 0; i < handles.length; i++) {
//    // if (alive) {
//    // handles[i].notifyObservers(NodeHandle.DECLARED_LIVE);
//    // } else {
//    // handles[i].notifyObservers(NodeHandle.DECLARED_DEAD);
//    // }
//    // }
//    // }
//  }


  public void registerIdentifier(Identifier i, TransportLayer<Identifier, MessageType> dtl, NodeRecord record) {
    nodes.put(i, new Tupel(i, dtl, record));
  }

  public void remove(Identifier i) {
    nodes.remove(i);
    notifyLivenessListeners(i, LivenessListener.LIVENESS_DEAD);
  }

  public Environment getEnvironment() {
    return environment;
  }

  
  public RandomSource getRandomSource() {
    return random;
  }

  public boolean isAlive(Identifier i) {
    return nodes.containsKey(i);
  }
  
  public DirectTransportLayer<Identifier, MessageType> getTL(Identifier i) {
    Tupel t = nodes.get(i);
    if (t == null) return null;
    NodeHandleAdapter nha = (NodeHandleAdapter)t.tl;
    return (DirectTransportLayer<Identifier, MessageType>)nha.getTL();
//    return (DirectTransportLayer<Identifier, MessageType>)nodes.get(i).tl;
  }
  
  /**
   * computes the one-way distance between two NodeIds
   * 
   * @param a the first NodeId
   * @param b the second NodeId
   * @return the proximity between the two input NodeIds
   */
  public float networkDelay(Identifier a, Identifier b) {
    Tupel ta = nodes.get(a);
    Tupel tb = nodes.get(b);
    if (ta == null) {
      throw new RuntimeException("asking about node proximity for unknown node "+a);
    }
      
    if (tb == null) {
      throw new RuntimeException("asking about node proximity for unknown node "+b);      
    }
    
    NodeRecord nra = ta.record;
    NodeRecord nrb = tb.record;
    
    return nra.networkDelay(nrb);
  }

  /**
   * computes the rtt between two NodeIds
   * 
   * @param a the first NodeId
   * @param b the second NodeId
   * @return the proximity between the two input NodeIds
   */
  public float proximity(Identifier a, Identifier b) {
    Tupel ta = nodes.get(a);
    Tupel tb = nodes.get(b);
    if (ta == null) {
      throw new RuntimeException("asking about node proximity for unknown node "+a);
    }
      
    if (tb == null) {
      throw new RuntimeException("asking about node proximity for unknown node "+b);      
    }
    
    NodeRecord nra = ta.record;
    NodeRecord nrb = tb.record;
    
    return nra.proximity(nrb);
  }

  public NodeRecord getNodeRecord(DirectNodeHandle handle) {
    Tupel t = nodes.get(handle);
    if (t == null) return null;
    return t.record;
  }

  List<LivenessListener<Identifier>> livenessListeners = new ArrayList<LivenessListener<Identifier>>();
  public void addLivenessListener(LivenessListener<Identifier> name) {
    synchronized(livenessListeners) {
      livenessListeners.add(name);
    }
  }

  public boolean removeLivenessListener(LivenessListener<Identifier> name) {
    synchronized(livenessListeners) {
      return livenessListeners.remove(name);
    }
  }
  
  private void notifyLivenessListeners(Identifier i, int liveness) {
    if (logger.level <= Logger.FINER) logger.log("notifyLivenessListeners("+i+","+liveness+"):"+livenessListeners.get(0));
    List<LivenessListener<Identifier>> temp;
    synchronized(livenessListeners) {
      temp = new ArrayList<LivenessListener<Identifier>>(livenessListeners);
    }
    for (LivenessListener<Identifier> listener : temp) {
      listener.livenessChanged(i, liveness);
    }
  }

  public boolean checkLiveness(Identifier i, Map<String, Integer> options) {
    return false;
  }

  public int getLiveness(Identifier i, Map<String, Integer> options) {
    if (nodes.containsKey(i)) {
      return LivenessListener.LIVENESS_ALIVE;
    }
    return LivenessListener.LIVENESS_DEAD;
  }

  public void clearState(Identifier i) {
    throw new IllegalStateException("not implemented");
  }
}
